import ast
from typing import Iterable, List, Optional, Sequence, TYPE_CHECKING, Union
from urllib.error import HTTPError
import webbrowser

from cytoolz import update_in
import rich.pretty

from grid.cli import globals as env
from grid.metadata import __version__
from grid.sdk.client import create_gql_client
from grid.sdk.client.grid_gql import gql_execute, gql_subscribe

if TYPE_CHECKING:
    from grid.sdk.experiments import ExperimentName

# TODO(rlizzo): standardize callers so that these all take ``Client`` instances.


def get_grid_user_id(client) -> str:
    query = """
    query Login ($cliVersion: String!) {
        cliLogin (cliVersion: $cliVersion) {
            userId
            success
            message
        }
    }
    """
    res = gql_execute(client, query, cliVersion=__version__)['cliLogin']
    return res['userId']


def get_user_basic_info():
    client = create_gql_client()
    query = """
    query {
        getUser {
            userId
            isVerified
            completedSignup
            isBlocked
            username
            firstName
            lastName
            email
        }
    }
    """
    return gql_execute(client, query)['getUser']


def get_experiment_details_gql(experiment_id: str):
    client = create_gql_client()
    query = """
    query ($experimentId: ID!) {
        getExperimentDetails (experimentId: $experimentId) {
            name
            experimentId
            githubId
            commitSha
            status
            desiredState
            createdAt
            startedRunningAt
            finishedAt
            invocationCommands
            parameters {
                name
                value
            }
            userDetails {
                username
                id
                firstName
                lastName
            }
        }
    }
    """
    resp = gql_execute(client, query, experimentId=experiment_id)
    res = resp['getExperimentDetails']
    _r1 = update_in(res, ['status'], str.upper)
    return update_in(_r1, ['desiredState'], str.upper)


def get_experiment_status_gql(experiment_id: str):
    client = create_gql_client()
    query = """
    query ($experimentId: ID!) {
        getExperimentDetails (experimentId: $experimentId) {
            status
            desiredState
            createdAt
            startedRunningAt
            finishedAt
        }
    }
    """
    res = gql_execute(client, query, experimentId=experiment_id)['getExperimentDetails']
    _r1 = update_in(res, ['status'], str.upper)
    return update_in(_r1, ['desiredState'], str.upper)


def cancel_experiment(exp_id: str) -> None:
    client = create_gql_client()
    query = """
        mutation ($experimentId: ID!) {
                cancelExperiment(experimentId: $experimentId) {
                    success
                    message
                }
            }
        """
    params = {'experimentId': exp_id}
    result = gql_execute(client, query, **params)['cancelExperiment']
    if result['success'] is not True:
        raise RuntimeError(result['message'])


def experiment_id_from_name(name: 'ExperimentName') -> str:
    """Retrieves experiment ID from an experiment name.

    Parameters
    ----------
    name
        experiment name struct which has been generated from
        parsing user code.

    Returns
    -------
    str
        The experiment ID to be used.
    """
    client = create_gql_client()
    query = """
    query ($experimentName: String!, $username: String) {
        getExperimentId(experimentName: $experimentName, username: $username) {
            success
            message
            experimentId
        }
    }
    """
    params = {"experimentName": name.name, "username": name.username}
    result = gql_execute(client, query, **params)
    if not result["getExperimentId"]["success"]:
        if "Cannot find experiment" in result["getExperimentId"]["message"]:
            raise KeyError(
                f"Experiment {name} does not exist\n "
                "If you wish to fetch an experiment for somebody in "
                "your team, use <username>:<experiment-name> format."
            )
        raise ValueError(f"{result['getExperimentId']['message']}")
    return result["getExperimentId"]["experimentId"]


def get_runs(run_name: Optional[str] = '', team_id: Optional[str] = None) -> Sequence[dict]:
    """Get the config for runs and shallow description of all contained experiments

    TODO(rlizzo):
        evaluate if all return fields of `experiment` in GQL query below
        are actually required. Simplify if possible

    Parameters
    ----------
    client
        gql client instance
    run_name
        Optional parameter to filter run config return to a specific run name.
        If not set, then all runs for the client's authenticated user are
        returned (optionally filtered by team if `team_id` filter param is set.
    team_id
        Optional parameter to filter run config results on the intersection
        of both the client's authenticated user AND a team ID.

    Returns
    -------
    Sequence[dict]
        An unstructured mapping of run config, directly maps to the query
        string which is contained in the implementation of this function.
    """
    client = create_gql_client()
    params = {'runName': run_name, "teamId": team_id}

    query = """
        query (
            $runName: ID, $teamId: ID
        ) {
            getRuns (runName: $runName, teamId: $teamId) {
                name
                runId
                projectId
                clusterId
                createdAt
                description
                entrypoint
                estimatedHourlyCost
                userDetails {
                    username
                    id
                    firstName
                    lastName
                }
                scriptCommand
                config {
                    compute
                    hyperParams
                }
                nExperiments
                experiments {
                    name
                    experimentId
                    commitSha
                    status
                    desiredState
                    createdAt
                    startedRunningAt
                    finishedAt
                    invocationCommands
                    parameters {
                        name
                        value
                    }
                }
            }
        }
        """
    result = gql_execute(client, query, **params)['getRuns']
    return result


def get_runs_lookup_data(run_name: Optional[str] = '', team_id: Optional[str] = None) -> Sequence[dict]:
    """Find run names (and associated id) for authenticated user.

    Parameters
    ----------
    client
        gql client instance
    run_name
        Optional parameter to filter run config return to a specific run name.
        If not set, then all runs for the client's authenticated user are
        returned (optionally filtered by team if `team_id` filter param is set.
    team_id
        Optional parameter to filter run config results on the intersection
        of both the client's authenticated user AND a team ID.

    Returns
    -------
    Sequence[dict]
        An unstructured mapping of run config, directly maps to the query
        string which is contained in the implementation of this function.
    """
    client = create_gql_client()
    query = """
        query (
            $runName: ID, $teamId: ID
        ) {
            getRuns (runName: $runName, teamId: $teamId) {
                name
                runId
            }
        }
        """
    params = {'runName': run_name, "teamId": team_id}
    result = gql_execute(client, query, **params)['getRuns']
    return result


def cancel_run(run_name: str) -> None:
    client = create_gql_client()
    query = """
        mutation ($name: ID!) {
            cancelRun(name: $name) {
                success
                message
            }
        }
        """
    result = gql_execute(client, query, name=run_name)['cancelRun']
    if result['success'] is not True:
        raise RuntimeError(result['message'])


def experiment_archived_build_logs(exp_id: str) -> Iterable[dict]:
    """Retrievies build logs from the archive.

    Yields
    ------
    Dict[str, str]
        Log lines
    """
    client = create_gql_client()
    query = """
    query ($experimentId: ID!) {
        getBuildLogs(experimentId: $experimentId) {
            message
            timestamp
        }
    }
    """
    params = {'experimentId': exp_id}
    result = gql_execute(client, query, **params)
    # The backend will return end empty object if no pages of logs are available.
    if not result.get('getBuildLogs'):
        raise StopIteration
    for log in result['getBuildLogs']:
        yield log


def experiment_live_build_logs(exp_id: str, limit: Union[int, None]) -> Iterable[dict]:
    """Streams build logs from subscription API.

    Parameters
    ----------
    exp_id
        experiment id which build logs should be retrieved for.
    limit
        if integer, terminate the subscription after N number of lines
        have been written to the client. If None, then this subscription
        never terminates.

    Yields
    ------
    Union[str, str]
        Log lines
    """
    client = create_gql_client(websocket=True)
    query = """
    subscription ($experimentId: ID!, $tailLines:Int) {
        getLiveBuildLogs(
            experimentId: $experimentId,
            tailLines: $tailLines) {
                message
                timestamp
        }
    }
    """
    params = {"experimentId": exp_id, "tailLines": limit}
    for msg in gql_subscribe(client, query, variable_values=params).get("getLiveBuildLogs"):
        yield msg['tailLines']


def experiment_archived_logs(exp_id: str) -> Iterable[dict]:
    client = create_gql_client()
    query = """
    query GetLogs ($experimentId: ID!) {
        getArchiveExperimentLogs(experimentId: $experimentId) {
            lines {
                message
                timestamp
            }
        }
    }
    """
    params = {'experimentId': exp_id}
    result = gql_execute(client, query, **params)

    # The backend will return end empty object if no pages of logs are available.
    if not result.get('getArchiveExperimentLogs'):
        raise StopIteration
    for log in result['getArchiveExperimentLogs']['lines']:
        yield log


def experiment_live_logs(exp_id: str, n_lines: int) -> Iterable[dict]:
    # If the experiment isn't in a finished state, then do a subscription with live logs.
    client = create_gql_client(websocket=True)
    query = """
    subscription GetLogs ($experimentId: ID!, $nLines: Int!) {
        getLiveExperimentLogs(
            experimentId: $experimentId, nLines: $nLines) {
                message
                timestamp
        }
    }
    """
    params = {'experimentId': exp_id, 'nLines': n_lines}
    stream = gql_subscribe(client, query, variable_values=params)
    for log in stream:
        yield log['getLiveExperimentLogs']


def _check_user_github_token() -> bool:
    """Checks if user has a valid Github token available.

    If user doesn't have one, then redirect user to the Grid UI to fetch a new one.

    Returns
    -------
    has_token: bool
        Boolean indicating if user has valid token.
    """
    # Build query
    client = create_gql_client()
    query = """
        query CheckToken {
            checkUserGithubToken {
                hasValidToken
            }
        }
    """

    # Check if the user has a token. If she hasn't, then redirect
    # user to the /auth page to get a new Github token.
    try:
        result = gql_execute(client, query)
        has_token = result['checkUserGithubToken']['hasValidToken']
        if not has_token:
            auth_url = env.GRID_URL.replace("graphql", "auth")
            webbrowser.open(auth_url, new=2, autoraise=True)
            raise PermissionError(
                "Authentication tokens need to be renewed! Opening Grid "
                "on the browser so we can renew your authentication tokens."
            )
    except HTTPError as e:
        raise PermissionError(e)

    return has_token


def repository_presigned_url(package_name: str, cluster_id: Optional[str] = None):
    """Pre-signed URL for uploading repository package to S3
    """
    client = create_gql_client()
    query = """
    query ($path: String!, $clusterId: ID) {
        getLocalDirPresignedUrl(path: $path, clusterId: $clusterId) {
            presignedUrl
        }
    }
    """
    params = {"path": package_name, "clusterId": cluster_id}
    result = gql_execute(client, query, **params)
    return result["getLocalDirPresignedUrl"]["presignedUrl"]


def check_github_repo_accessible(github_repository: str) -> bool:
    """
    Checks if user has authoirized Grid to access a given Github repository.
    If the user did not, then redirect them to the
    Grid UI settings page to authorize.

    Parameters
    ----------
    github_repository: str
        Github repository URL or ID

    Returns
    -------
    is_accessible: bool
        Boolean indicating if the user authorized to access the Github repository,
    """
    client = create_gql_client()
    query = """
        query ($repositoryUrl: ID!) {
            isGithubRepositoryAccessible(repositoryUrl: $repositoryUrl) {
                isAccessible
            }
        }
    """
    result = gql_execute(client, query, repositoryUrl=github_repository)
    return result['isGithubRepositoryAccessible']['isAccessible']


def start_run_mutation(
    script_args, entrypoint, config_encoded, run_name, run_description, commit_sha, github_repository,
    invocation_command
):
    # Build GraphQL query
    client = create_gql_client()
    mutation = """
    mutation (
        $configString: String!
        $name: String!
        $description: String
        $commitSha: String
        $githubRepository: ID!
        $commandLineArgs: [String]!
        $invocationCommand: String!
        ) {
        trainScript (
            properties: {
                    githubRepository: $githubRepository
                    name: $name
                    description: $description
                    configString: $configString
                    commitSha: $commitSha
                    commandLineArgs: $commandLineArgs
                    invocationCommand: $invocationCommand
                }
        ) {
        success
        message
        name
        runId
        }
    }
    """

    #  Prepend the file name to the list of args and builds the query payload.
    script_args.insert(0, str(entrypoint))
    params = {
        'configString': config_encoded,
        'name': run_name,
        'description': run_description,
        'commitSha': commit_sha,
        'githubRepository': str(github_repository),
        'commandLineArgs': script_args,
        'invocationCommand': invocation_command
    }

    #  Send request to Grid.
    try:
        result = gql_execute(client, mutation, **params)
    except Exception as e:
        rich.pretty.pprint(e)
        message = ast.literal_eval(str(e))['message']
        CREDIT_CARD_ERROR_MESSAGE = "A credit card on file is needed in order to use a GPU"
        if CREDIT_CARD_ERROR_MESSAGE in message:
            PermissionError(message)
        raise


def estimate_run_cost_mutation(script_args, entrypoint, config_encoded):
    # Build GraphQL query
    client = create_gql_client()
    mutation = """
    mutation (
        $configString: String!
        $commandLineArgs: [String]!
        ) {
        estimateTrainScript (
            properties: {
                    configString: $configString
                    commandLineArgs: $commandLineArgs
                }
        ) {
        success
        message
        nExperiments
        estimatedHourlyCost
        }
    }
    """

    #  Prepend the file name to the list of args and builds the query payload.
    script_args.insert(0, str(entrypoint))
    params = {
        'configString': config_encoded,
        'commandLineArgs': script_args,
    }

    #  Send request to Grid.
    try:
        result = gql_execute(client, mutation, **params)
    except Exception as e:
        message = ast.literal_eval(str(e))['message']
        CREDIT_CARD_ERROR_MESSAGE = "A credit card on file is needed in order to use a GPU"
        if CREDIT_CARD_ERROR_MESSAGE in message:
            PermissionError(message)
        raise

    return result['estimateTrainScript']


def get_user_teams() -> List[dict]:
    client = create_gql_client()
    query = """
        query GetUserTeams {
            getUserTeams {
                success
                message
                teams {
                    id
                    name
                    createdAt
                    role
                    members {
                        id
                        username
                        firstName
                        lastName
                    }
                }
            }
        }
    """
    result = gql_execute(client, query)
    if not result['getUserTeams'] or not result['getUserTeams']['success']:
        raise RuntimeError(result['getUserTeams']["message"])
    return result['getUserTeams']['teams']


def get_available_datastores(team_id: Optional[str] = None):
    client = create_gql_client()
    query = """
        query GetDatastores ($teamId: ID){
            getDatastores(teamId: $teamId) {
                id
                name
                version
                size
                createdAt
                snapshotStatus
                clusterId
                userDetails {
                    id
                    username
                    firstName
                    lastName
                }
            }
        }
    """
    params = {'teamId': team_id}
    result = gql_execute(client, query, **params)
    return result['getDatastores']


def delete_datastore(name: str, version: int, cluster: Optional[str] = None):
    """Delete datastore for user

    Parameters
    ----------
    name
        Datastore name
    version
        Datastore version
    cluster
        cluster id to operate on

    Raises
    ------
    RuntimeError
        If datastore deletion fails
    """
    client = create_gql_client()
    mutation = """
        mutation (
            $name: String!
            $version: Int!
            $clusterId: String
            ) {
            deleteDatastore (
                properties: {
                        name: $name,
                        version: $version,
                        clusterId: $clusterId
                    }
            ) {
            success
            message
            }
        }
    """
    params = {'name': name, 'version': version, 'clusterId': cluster}
    result = gql_execute(client, mutation, **params)

    if not result['deleteDatastore']['success']:
        message = result['deleteDatastore']['message']
        raise RuntimeError(f'failed to delete datastore {name} with version {version}: {message}')


def create_datastore(name: str, source: str, cluster: Optional[str] = None):
    """Create datastore in Grids
    """
    # Create Grid datastore directly in Grid without uploading, since Grid will
    # handle extraction and creating a optimizted datastore automatically.
    client = create_gql_client()
    mutation = """
        mutation (
            $name: String!
            $source: String
            $clusterId: String
            ) {
            createDatastore (
                properties: {
                        name: $name
                        source: $source
                        clusterId: $clusterId
                    }
            ) {
            success
            message
            datastoreId
            datastoreVersion
            }
        }
    """

    params = {'name': name, 'source': source, 'clusterId': cluster}
    result = gql_execute(client, mutation, **params)
    success = result['createDatastore']['success']
    message = result['createDatastore']['message']
    if not success:
        raise ValueError(f"Unable to create datastore: {message}")

    res = result['createDatastore']
    res['datastoreVersion'] = int(res['datastoreVersion'])
    return res


def get_datastore_upload_multipart_presigned_urls(path: str, datastore_id: str, count: int):
    client = create_gql_client()
    query = """
        query GetMultiPartPresignedUrls (
            $path: String!,
            $datastoreId: ID!,
            $count: Int!
        ) {
            getMultiPartPresignedUrls (
                path: $path,
                datastoreId: $datastoreId,
                count: $count
            ) {
                uploadId
                presignedUrls {
                    url
                    part
                }
            }
        }
    """
    params = {
        'path': path,
        'count': count,
        'datastoreId': datastore_id,
    }
    result = gql_execute(client, query, **params)
    return result['getMultiPartPresignedUrls']


def complete_multipart_datastore_upload(datastore_id: str, upload_id: str, parts: str, path: str, cluster_id: str):
    client = create_gql_client()
    mutation = """
         mutation (
             $datastoreId: ID!
             $uploadId: String!
             $parts: JSONString!
             $path: String!
             $clusterId: String
             ) {
             completeMultipartDatastoreUpload (
                 properties: {
                         datastoreId: $datastoreId
                         uploadId: $uploadId
                         parts: $parts
                         path: $path
                         clusterId: $clusterId
                     }
             ) {
             success
             message
             }
         }
     """
    params = {
        'datastoreId': datastore_id,
        'uploadId': upload_id,
        'parts': parts,
        'path': path,
        'clusterId': cluster_id,
    }
    result = gql_execute(client, mutation, **params)
    success = result['completeMultipartDatastoreUpload']['success']
    message = result['completeMultipartDatastoreUpload']['message']
    if not success:
        raise ValueError(f"Unable to complete multi-part upload: {message}")


def complete_datastore_upload(datastore_id: str, size_mib: int):
    client = create_gql_client()
    mutation = """
        mutation (
            $datastoreId: ID!
            $sizeMib: Int!
            ) {
            completeDatastoreUpload (
                properties: {
                        datastoreId: $datastoreId
                        sizeMib: $sizeMib
                    }
            ) {
            success
            message
            }
        }
    """
    params = {'datastoreId': datastore_id, 'sizeMib': size_mib}
    result = gql_execute(client, mutation, **params)

    success = result['completeDatastoreUpload']['success']
    message = result['completeDatastoreUpload']['message']
    if not success:
        raise ValueError(f"Unable to complete datastore upload: {message}")


def get_user():
    client = create_gql_client()
    query = """
        query {
            getUser {
                isVerified
                completedSignup
                isBlocked
            }
        }
        """
    result = gql_execute(client, query)
    return result["getUser"]



"""This module complements missing features of ``numpy.random``.

The module contains:
    * Several algorithms to sample integers without replacement.
"""
# Attribution
# -----------
#
# This code has been adapted from the scikit-learn project. Details of original source
# material and how to access are provided below:
#
# - Repository:    https://github.com/scikit-learn/scikit-learn
# - Commit:        ed3642014be412b0bda13d1ec756baeabb0dcbfe
# - File Path:     sklearn/model_selection/_search.py
#                  sklearn/utils/_random.pyx
#                  sklearn/utils/random.py
# - License:       BSD 3 Clause
# - Access Date:   11 July 2021
# - Adapted By:    Richard Izzo (rlizzo)
# - License Text:  Please refer to NOTICE file in this repository for full
#                  text of original source license.
import numbers
from typing import Optional, Union

import numpy as np


def _sample_without_replacement_check_input(n_population: int, n_samples: int):
    """Check that input are consistent for sample_without_replacement.
    """
    if n_population < 0:
        raise ValueError(f'n_population should be greater than 0, got {n_population}.')

    if n_samples > n_population:
        raise ValueError(
            f'n_population should be greater or equal than n_samples, got '
            f'n_samples > n_population ({n_samples} > {n_population})'
        )


def _sample_without_replacement_with_tracking_selection(n_population: int, n_samples: int, random_state=None):
    r"""Sample integers without replacement.

    Select n_samples integers from the set [0, n_population) without
    replacement.

    Time complexity:
        - Worst-case: unbounded
        - Average-case:
            O(O(np.random.randint) * \sum_{i=1}^n_samples 1 /
                                              (1 - i / n_population)))
            <= O(O(np.random.randint) *
                   n_population * ln((n_population - 2)
                                     /(n_population - 1 - n_samples)))
            <= O(O(np.random.randint) *
                 n_population * 1 / (1 - n_samples / n_population))

    Space complexity of O(n_samples) in a python set.

    Parameters
    ----------
    n_population : int
        The size of the set to sample from.

    n_samples : int
        The number of integer to sample.

    random_state : int, RandomState instance or None, default=None
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    Returns
    -------
    out : ndarray of shape (n_samples,)
        The sampled subsets of integer.
    """
    _sample_without_replacement_check_input(n_population, n_samples)

    out = np.empty((n_samples, ), dtype=int)  # cdef np.ndarray[np.int_t, ndim=1]

    rng = check_random_state(random_state)
    rng_randint = rng.randint

    # The following line of code are heavily inspired from python core,
    # more precisely of random.sample.
    selected = set()  # cdef set

    for i in range(n_samples):
        j = rng_randint(n_population)
        while j in selected:
            j = rng_randint(n_population)
        selected.add(j)
        out[i] = j

    return out


def _sample_without_replacement_with_pool(n_population: int, n_samples: int, random_state=None):
    """Sample integers without replacement.

    Select n_samples integers from the set [0, n_population) without
    replacement.

    Time complexity: O(n_population +  O(np.random.randint) * n_samples)

    Space complexity of O(n_population + n_samples).

    Parameters
    ----------
    n_population : int
        The size of the set to sample from.

    n_samples : int
        The number of integer to sample.

    random_state : int, RandomState instance or None, default=None
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    Returns
    -------
    out : ndarray of shape (n_samples,)
        The sampled subsets of integer.
    """
    _sample_without_replacement_check_input(n_population, n_samples)

    out = np.empty((n_samples, ), dtype=int)  # cdef np.ndarray[np.int_t, ndim=1]
    pool = np.empty((n_population, ), dtype=int)  # cdef np.ndarray[np.int_t, ndim=1]

    rng = check_random_state(random_state)
    rng_randint = rng.randint

    # Initialize the pool
    for i in range(n_population):
        pool[i] = i

    # The following line of code are heavily inspired from python core,
    # more precisely of random.sample.
    for i in range(n_samples):
        j = rng_randint(n_population - i)  # invariant: non-selected at [0,n-i)
        out[i] = pool[j]
        pool[j] = pool[n_population - i - 1]  # move non-selected item into vacancy

    return out


def _sample_without_replacement_with_reservoir_sampling(n_population: int, n_samples: int, random_state=None):
    """Sample integers without replacement.

    Select n_samples integers from the set [0, n_population) without
    replacement.

    Time complexity of
        O((n_population - n_samples) * O(np.random.randint) + n_samples)

    Space complexity of O(n_samples)

    Parameters
    ----------
    n_population : int
        The size of the set to sample from.

    n_samples : int
         The number of integer to sample.

    random_state : int, RandomState instance or None, default=None
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    Returns
    -------
    out : ndarray of shape (n_samples,)
        The sampled subsets of integer. The order of the items is not
        necessarily random. Use a random permutation of the array if the order
        of the items has to be randomized.
    """
    _sample_without_replacement_check_input(n_population, n_samples)

    out = np.empty((n_samples, ), dtype=int)  # cdef np.ndarray[np.int_t, ndim=1]

    rng = check_random_state(random_state)
    rng_randint = rng.randint

    # This cython implementation is based on the one of Robert Kern:
    # http://mail.scipy.org/pipermail/numpy-discussion/2010-December/
    # 054289.html
    #
    for i in range(n_samples):
        out[i] = i

    for i in range(n_samples, n_population):  # for i from n_samples <= i < n_population:
        j = rng_randint(0, i + 1)
        if j < n_samples:
            out[j] = i
    return out


def sample_without_replacement(n_population: int, n_samples: int, method: str = "auto", random_state=None):
    """Sample integers without replacement.

    Select n_samples integers from the set [0, n_population) without
    replacement.


    Parameters
    ----------
    n_population : int
        The size of the set to sample from.

    n_samples : int
        The number of integer to sample.

    random_state : int, RandomState instance or None, default=None
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    method : {"auto", "tracking_selection", "reservoir_sampling", "pool"}, \
            default='auto'
        If method == "auto", the ratio of n_samples / n_population is used
        to determine which algorithm to use:
        If ratio is between 0 and 0.01, tracking selection is used.
        If ratio is between 0.01 and 0.99, numpy.random.permutation is used.
        If ratio is greater than 0.99, reservoir sampling is used.
        The order of the selected integers is undefined. If a random order is
        desired, the selected subset should be shuffled.

        If method =="tracking_selection", a set based implementation is used
        which is suitable for `n_samples` <<< `n_population`.

        If method == "reservoir_sampling", a reservoir sampling algorithm is
        used which is suitable for high memory constraint or when
        O(`n_samples`) ~ O(`n_population`).
        The order of the selected integers is undefined. If a random order is
        desired, the selected subset should be shuffled.

        If method == "pool", a pool based algorithm is particularly fast, even
        faster than the tracking selection method. However, a vector containing
        the entire population has to be initialized.
        If n_samples ~ n_population, the reservoir sampling method is faster.

    Returns
    -------
    out : ndarray of shape (n_samples,)
        The sampled subsets of integer. The subset of selected integer might
        not be randomized, see the method argument.
    """
    _sample_without_replacement_check_input(n_population, n_samples)
    all_methods = ("auto", "tracking_selection", "reservoir_sampling", "pool")
    ratio = n_samples / n_population if n_population != 0.0 else 1.0

    # Check ratio and use permutation unless ratio < 0.01 or ratio > 0.99
    if method == "auto" and 0.01 < ratio < 0.99:
        rng = check_random_state(random_state)
        return rng.permutation(n_population)[:n_samples]

    if method == "auto" or method == "tracking_selection":
        # TODO the pool based method can also be used.
        #      however, it requires special benchmark to take into account
        #      the memory requirement of the array vs the set.
        # The value 0.2 has been determined through benchmarking.
        if ratio < 0.2:
            return _sample_without_replacement_with_tracking_selection(n_population, n_samples, random_state)
        else:
            return _sample_without_replacement_with_reservoir_sampling(n_population, n_samples, random_state)
    elif method == "reservoir_sampling":
        return _sample_without_replacement_with_reservoir_sampling(n_population, n_samples, random_state)
    elif method == "pool":
        return _sample_without_replacement_with_pool(n_population, n_samples, random_state)
    else:
        raise ValueError(f'Expected a method name in {all_methods}, got {method}.')


def check_random_state(seed: Optional[Union['numbers.Integral', 'np.random.RandomState']]):
    """Turn seed into a np.random.RandomState instance

    Parameters
    ----------
    seed :
        If seed is None, return the RandomState singleton used by np.random.
        If seed is an int, return a new RandomState instance seeded with seed.
        If seed is already a RandomState instance, return it.
        Otherwise raise ValueError.
    """
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, numbers.Integral):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError(f"{seed} cannot be used to seed a numpy.random.RandomState instance")



from grid.sdk.api import (
    cancel_experiment,
    cancel_run,
    create_run,
    Datastore,
    get_build_logs,
    get_experiment_by_id,
    get_experiment_by_name,
    get_experiment_logs,
    get_instance_types,
    get_run,
    list_clusters,
    list_datastores,
    list_runs,
    login,
    Run,
)

__all__ = [
    "login", "create_run", "list_runs", "list_datastores", "get_run", "cancel_run", "get_experiment_by_name",
    "get_experiment_by_id", "get_build_logs", "get_experiment_logs", "cancel_experiment", "Run", "Datastore",
    "get_instance_types", "list_clusters"
]



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


import pprint
import re  # noqa: F401

import six

from grid.openapi.configuration import Configuration


class Externalv1UpdateSessionResponse(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'cost': 'float',
        'created_at': 'datetime',
        'hourly_cost': 'float',
        'id': 'str',
        'name': 'str',
        'spec': 'V1SessionSpec',
        'status': 'V1SessionStatus'
    }

    attribute_map = {
        'cost': 'cost',
        'created_at': 'createdAt',
        'hourly_cost': 'hourlyCost',
        'id': 'id',
        'name': 'name',
        'spec': 'spec',
        'status': 'status'
    }

    def __init__(self, cost=None, created_at=None, hourly_cost=None, id=None, name=None, spec=None, status=None, _configuration=None):  # noqa: E501
        """Externalv1UpdateSessionResponse - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._cost = None
        self._created_at = None
        self._hourly_cost = None
        self._id = None
        self._name = None
        self._spec = None
        self._status = None
        self.discriminator = None

        if cost is not None:
            self.cost = cost
        if created_at is not None:
            self.created_at = created_at
        if hourly_cost is not None:
            self.hourly_cost = hourly_cost
        if id is not None:
            self.id = id
        if name is not None:
            self.name = name
        if spec is not None:
            self.spec = spec
        if status is not None:
            self.status = status

    @property
    def cost(self):
        """Gets the cost of this Externalv1UpdateSessionResponse.  # noqa: E501


        :return: The cost of this Externalv1UpdateSessionResponse.  # noqa: E501
        :rtype: float
        """
        return self._cost

    @cost.setter
    def cost(self, cost):
        """Sets the cost of this Externalv1UpdateSessionResponse.


        :param cost: The cost of this Externalv1UpdateSessionResponse.  # noqa: E501
        :type: float
        """

        self._cost = cost

    @property
    def created_at(self):
        """Gets the created_at of this Externalv1UpdateSessionResponse.  # noqa: E501


        :return: The created_at of this Externalv1UpdateSessionResponse.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this Externalv1UpdateSessionResponse.


        :param created_at: The created_at of this Externalv1UpdateSessionResponse.  # noqa: E501
        :type: datetime
        """

        self._created_at = created_at

    @property
    def hourly_cost(self):
        """Gets the hourly_cost of this Externalv1UpdateSessionResponse.  # noqa: E501


        :return: The hourly_cost of this Externalv1UpdateSessionResponse.  # noqa: E501
        :rtype: float
        """
        return self._hourly_cost

    @hourly_cost.setter
    def hourly_cost(self, hourly_cost):
        """Sets the hourly_cost of this Externalv1UpdateSessionResponse.


        :param hourly_cost: The hourly_cost of this Externalv1UpdateSessionResponse.  # noqa: E501
        :type: float
        """

        self._hourly_cost = hourly_cost

    @property
    def id(self):
        """Gets the id of this Externalv1UpdateSessionResponse.  # noqa: E501


        :return: The id of this Externalv1UpdateSessionResponse.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this Externalv1UpdateSessionResponse.


        :param id: The id of this Externalv1UpdateSessionResponse.  # noqa: E501
        :type: str
        """

        self._id = id

    @property
    def name(self):
        """Gets the name of this Externalv1UpdateSessionResponse.  # noqa: E501


        :return: The name of this Externalv1UpdateSessionResponse.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this Externalv1UpdateSessionResponse.


        :param name: The name of this Externalv1UpdateSessionResponse.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def spec(self):
        """Gets the spec of this Externalv1UpdateSessionResponse.  # noqa: E501


        :return: The spec of this Externalv1UpdateSessionResponse.  # noqa: E501
        :rtype: V1SessionSpec
        """
        return self._spec

    @spec.setter
    def spec(self, spec):
        """Sets the spec of this Externalv1UpdateSessionResponse.


        :param spec: The spec of this Externalv1UpdateSessionResponse.  # noqa: E501
        :type: V1SessionSpec
        """

        self._spec = spec

    @property
    def status(self):
        """Gets the status of this Externalv1UpdateSessionResponse.  # noqa: E501


        :return: The status of this Externalv1UpdateSessionResponse.  # noqa: E501
        :rtype: V1SessionStatus
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this Externalv1UpdateSessionResponse.


        :param status: The status of this Externalv1UpdateSessionResponse.  # noqa: E501
        :type: V1SessionStatus
        """

        self._status = status

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(Externalv1UpdateSessionResponse, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Externalv1UpdateSessionResponse):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, Externalv1UpdateSessionResponse):
            return True

        return self.to_dict() != other.to_dict()
